# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Date: 04.09.2016                                                            #
# Author: Ole-Johan Skrede                                                    #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
Top script for solution proposal of exercise 1.
"""

# pylint: disable=expression-not-assigned
# pylint: disable=bad-indentation

# Put standard library imports here
import os

# Put related third party imports here
import cv2
import matplotlib.pyplot as plt

# Put your own local imports here
import image_preprocessing
import texture

def plot_image(image, fig_num=0, title='Image', colormap='gray',
               colorbar=False, aspect='auto'):
  """Plot function for images"""

  if not colorbar:
    plt.figure(fig_num)
    plt.imshow(image, cmap=colormap, interpolation='none', aspect=aspect)
    plt.title(title)
    plt.xticks([])
    plt.yticks([])
  else:
    # vertical colorbar
    fig, ax = plt.subplots()
    cax = ax.imshow(image, cmap=colormap, interpolation='none', aspect=aspect)
    ax.set_title(title)
    fig.colorbar(cax)
  fig_num += 1

  return fig_num

def plot_image_and_histogram(image, name, fig_num):
  """Plot an image and its histogram on top of each other."""

  plt.figure(fig_num)
  plt.suptitle(name)

  plt.subplot(2, 1, 1)
  plt.imshow(image, 'gray'), plt.title('Image'), plt.xticks([]), plt.yticks([])

  plt.subplot(2, 1, 2)
  plt.hist(image.ravel(), 256), plt.title('Gray level histogram')

  fig_num += 1
  return fig_num


def compute_glcm_features(pp_image, feature_list, feature_dir, window_order,
                          delta_x, delta_y, isotropic_glcm, feature_postfix):
  """Compute glcm features from pp_image based on feature_list and parameters,
     and write them.

  Args:
    pp_image: 2D numpy array. Input image that is to be analyzed. Ideally
              preprocessed to reduce number of graylevels.
    feature_list: List of strings. List of names of the features we want to
                  compute.
    feature_dir: String. Where to write the feature results (stored as images).
    window_order: Int. 'Radius' of a square neighbourhood around some center
                  pixel.
    delta_x: Int. Vertical distance between compared pixels.
    delta_y: Int. Horizontal distance between compared pixels.
    isotropic_glcm: Bool. Whether or not to compute isotropic glcm.
    feature_postfix: String. Signature used to identify the different
                     parameters that are similar across features, for this run.
  """

  # Remove features from feature_list if they already have been computed, and
  # written
  if not os.path.exists(feature_dir):
    os.makedirs(feature_dir)
  existing_files = os.listdir(feature_dir)
  feature_list_copy = feature_list[:]
  for feature_name in feature_list_copy:
    # The filenames for the features should follow this convention
    filename = feature_name + "_image_" + feature_postfix + ".png"
    if filename in existing_files:
      feature_list.remove(feature_name)

  # Construct feature images from glcm analysis
  tex = texture.Texture(pp_image, window_order, delta_x, delta_y, isotropic_glcm)
  tex.run_texture_analysis(feature_postfix, feature_list=feature_list)

  for name, feature in tex.feature_images.items():
    feature.scale_image(0, 255)
    #fig_num = plot_image(feature.image, fig_num=fig_num, title=name,
    #                     colormap='cubehelix', colorbar=True)

    cv2.imwrite(os.path.join(feature_dir, name + '_image_' + feature.postfix + '.png'),
                feature.image)

def second_order_segmentation(image, feature_dir, result_dir, feature_name,
                              feature_postfix, threshold, above=True, fig_num=0):
  """Segmentation of an image based on a glcm feature, and a threshold"""
  feature_name = feature_name + "_image_" + feature_postfix
  feature_filename = os.path.join(feature_dir, feature_name + ".png")
  feature_image = cv2.imread(feature_filename, cv2.IMREAD_GRAYSCALE)
  if image.shape != feature_image.shape:
    feature_image = image_preprocessing.resize_image(feature_image,
                                                     image.shape[0], image.shape[1])

  mask = (feature_image < threshold)*(1 - above*1) + (feature_image >= threshold)*(above*1)

  m_image = image*mask

  fig_num = plot_image(feature_image, fig_num=fig_num, title=feature_name,
                       colormap='cubehelix', colorbar=True)
  fig_num = plot_image(m_image, fig_num=fig_num, title=feature_name + " mask")

  cv2.imwrite(os.path.join(result_dir, feature_name + "_feature.png"), feature_image)
  cv2.imwrite(os.path.join(result_dir, feature_name + "_masked.png"), m_image)

  return fig_num

def first_order_segmentation(image, window_order, result_dir,
                             threshold_feature, threshold, above=True, fig_num=0):
  """Segmentation of an image based on first order texture analysis, and a threshold"""

  if threshold_feature == 'variance':
    # First order variance
    var_image = texture.first_order_variance(image, window_order)
    var_image = image_preprocessing.scale_image(var_image, 0, 255)
    var_mask = (var_image < threshold)*(1 - above*1) + (var_image >= threshold)*(above*1)
    var_masked = image*var_mask
    fig_num = plot_image(var_image, fig_num=fig_num, title='First order variance',
                         colormap='cubehelix', colorbar=True)
    fig_num = plot_image(var_masked, fig_num=fig_num, title='Variance masked')

    cv2.imwrite(os.path.join(result_dir,
                             "first_order_variance_" + str(window_order) + "_feature.png"),
                var_image)
    cv2.imwrite(os.path.join(result_dir,
                             "first_order_variance_" + str(window_order) + "_masked.png"),
                var_masked)

  if threshold_feature == 'entropy':
    # First order entropy
    ent_image = texture.first_order_entropy(image, window_order)
    ent_image = image_preprocessing.scale_image(ent_image, 0, 255)
    ent_mask = (ent_image < threshold)*(1 - above*1) + (ent_image >= threshold)*(above*1)
    ent_masked = image*ent_mask
    fig_num = plot_image(ent_image, fig_num=fig_num,
                         title='First order entropy, wo=' + str(window_order),
                         colormap='cubehelix', colorbar=True)
    fig_num = plot_image(ent_masked, fig_num=fig_num,
                         title='Entropy masked, wo=' + str(window_order))

    cv2.imwrite(os.path.join(result_dir,
                             "first_order_entropy_" + str(window_order) + "_feature.png"),
                ent_image)
    cv2.imwrite(os.path.join(result_dir,
                             "first_order_entropy_" + str(window_order) + "_masked.png"),
                ent_masked)

  return fig_num

def laws_segmentation(image, result_dir, threshold, above=True, fig_num=0):
  """Laws energy features used to analyze an image."""

  laws_features = texture.laws_energy_features(image)

  for name, feat_im in laws_features.items():
    feat_im = image_preprocessing.scale_image(feat_im, 0, 255)
    mask = (feat_im < threshold)*(1 - above*1) + (feat_im >= threshold)*(above*1)
    masked_im = image*mask
    fig_num = plot_image(feat_im, fig_num=fig_num, title=name,
                         colormap='cubehelix', colorbar=True)
    fig_num = plot_image(masked_im, fig_num=fig_num, title='Masked from '+name)

    cv2.imwrite(os.path.join(result_dir, "laws_" + name + "_feature.png"), feat_im)
    cv2.imwrite(os.path.join(result_dir, "laws_" + name + "_masked.png"), masked_im)

  return fig_num


def main():
  """Run this for solution to task 1, 2, 3 and 4"""

  print('='*80)
  print('Solution to weekly exercises in INF4300')
  print('Exercise 2')
  print('-'*80)

  # Configurations

  image_dir = '../../images'
  image_file = 'zebra_1.tif'
  image_name = os.path.splitext(image_file)[0]
  full_image_path = os.path.join(image_dir, image_file)
  feature_dir = './features/' + image_name
  result_dir = './results'

  new_height = None # Set to None if you want to use original size
  new_width = None # Set to None if you want to use original size

  num_greylevels = 8
  window_order = 10
  delta_x = 10
  delta_y = -10
  isotropic_glcm = False
  feature_postfix = 'lvl{0}_win{1}_dx{2}_dy{3}_{4}isotropy'.format(
      num_greylevels,
      window_order,
      delta_x,
      delta_y,
      '' if isotropic_glcm else 'an')

  # The time of computing extra features is dwarfed by the time computing the
  # glcm. So we add a bit more.
  feature_list = ['max_prob',
                  'entropy',
                  'uniformity',
                  'homogeneity',
                  'glcm_var',
                  'inertia', # contrast
                  'correlation',
                  'cluster_shade'
                 ]

  tasks = ['task1'] # List of tasks you want to execute
  write_results = False

  threshold_feature = 'inertia'
  threshold = 40
  above = True




  fig_num = 0

  # Read the original image.
  image = cv2.imread(full_image_path, cv2.IMREAD_GRAYSCALE)
  fig_num = plot_image(image, fig_num=fig_num, title='Original')

  pp_image = image_preprocessing.preprocess(image, num_greylevels, new_height,
                                            new_width)
  fig_num = plot_image(pp_image, fig_num=fig_num, title='Preprocessed')

  if write_results:
    cv2.imwrite(os.path.join(result_dir, image_name + '_original.png'), image)
    cv2.imwrite(os.path.join(result_dir, image_name + '_preprocessed.png'), pp_image)

  #print('Preprocessed image info')
  #print('Shape: ', pp_image.shape)
  #print('Data type: ', pp_image.dtype)

  if 'task1' in tasks:
    # Compute features (if they are not already computed and written)
    compute_glcm_features(pp_image, feature_list, feature_dir, window_order,
                          delta_x, delta_y, isotropic_glcm, feature_postfix)

  if 'task2' in tasks:
    # Create mask from image segmentation based on glcm features
    fig_num = second_order_segmentation(image, feature_dir, result_dir,
                                        threshold_feature, feature_postfix,
                                        threshold, above, fig_num)

  if 'task3' in tasks:
    # Create mask from image segmentation based on first order features
    fig_num = first_order_segmentation(image, window_order, result_dir,
                                       threshold_feature, threshold, above,
                                       fig_num)

  if 'task4' in tasks:
    # Create mask from image segmentation based on Laws energy features
    fig_num = laws_segmentation(image, result_dir, threshold, above, fig_num)

  plt.show()

if __name__ == "__main__":
  main()
